@@ -7,8 +7,7 @@
 use crate::document::FileInfo;
 use crate::event::{Error, Result, Status};
 use crate::st;
-use crate::utils::raw_indices;
-use unicode_width::{UnicodeWidthChar, UnicodeWidthStr};
+use crate::utils::{raw_indices, width, width_char, BoundedRange};
 
 /// A struct that contains all the basic tools necessary to manage rows in a document
 #[derive(Debug, PartialEq, Clone)]
@@ -54,7 +53,9 @@ impl Row {
     pub fn link(mut self, link: *mut FileInfo) -> Self {
         self.info = link;
         let tabs = self.get_tab_width();
-        self.indices = Row::raw_to_indices(&self.text, tabs);
+        if tabs != 4 {
+            self.indices = Row::raw_to_indices(&self.text, tabs);
+        }
         self
     }
 
@@ -74,20 +75,26 @@ impl Row {
     }
 
     /// Remove text in a range
+    ///
+    /// This takes in an inclusive or exclusive range: `..` and `..=` only.
     /// # Errors
     /// Will return `Err` if `range` is out of range of the row
+    #[allow(clippy::needless_pass_by_value)]
     pub fn remove<R>(&mut self, range: R) -> Result<Status>
     where
-        R: std::ops::RangeBounds<usize>,
+        R: BoundedRange,
     {
-        if let std::ops::Bound::Included(start) = range.start_bound() {
-            if start > &self.width() {
-                return Err(Error::OutOfRange);
-            }
+        let (start, end) = (range.first(), range.last());
+        if start > self.width() {
+            return Err(Error::OutOfRange);
         }
-        self.text.splice(range, []);
-        let tabs = self.get_tab_width();
-        self.indices = Row::raw_to_indices(&self.text, tabs);
+        let shift = self.indices[end] - self.indices[start];
+        self.text.splice(start..end, []);
+        self.indices.splice(start..end, []);
+        self.indices
+            .iter_mut()
+            .skip(start)
+            .for_each(|i| *i -= shift);
         self.modified = true;
         Ok(Status::None)
     }
@@ -109,16 +116,20 @@ impl Row {
             modified: false,
         };
         // Shift down
-        let shift = *right.indices.first().unwrap();
+        let shift = *right.indices.first().unwrap_or(&0);
         right.indices.iter_mut().for_each(|i| *i -= shift);
         Ok((left, right))
     }
 
     /// Joins this row with another row
     pub fn splice(&mut self, mut row: Row) -> Row {
+        let mut indices = self.indices.clone();
+        let shift = *self.indices.last().unwrap_or(&0);
+        row.indices.remove(0);
+        row.indices.iter_mut().for_each(|i| *i += shift);
+        indices.append(&mut row.indices);
         let mut text = self.text.clone();
         text.append(&mut row.text);
-        let indices = Row::raw_to_indices(&text, self.get_tab_width());
         Row {
             indices,
             text,
@@ -131,17 +142,45 @@ impl Row {
     /// ```
     /// // Opening a file,
     /// use kaolinite::document::Document;
-    /// let mut doc = Document::new();
+    /// let mut doc = Document::new((10, 10));
     /// // Imagine if test.txt were `The quick brown fox`
-    /// doc.open("test.txt").expect("Failed to open file");
+    /// doc.open("examples/test.txt").expect("Failed to open file");
     /// // This would get the word boundaries of the first row: [0, 4, 10, 16, 19]
-    /// println!("{:?}", doc.row(0).words());
+    /// println!("{:?}", doc.row(0).unwrap().words());
     /// ```
     #[must_use]
     pub fn words(&self) -> Vec<usize> {
         crate::utils::words(self)
     }
 
+    /// Find the next word in this row from the character index
+    #[must_use]
+    pub fn next_word_forth(&self, loc: usize) -> usize {
+        let bounds = self.words();
+        let mut last = *bounds.last().unwrap_or(&0);
+        for bound in bounds.iter().rev() {
+            if bound <= &loc {
+                return last;
+            }
+            last = *bound;
+        }
+        unreachable!()
+    }
+
+    /// Find the previous word in this row from the character index
+    #[must_use]
+    pub fn next_word_back(&self, loc: usize) -> usize {
+        let bounds = self.words();
+        let mut last = 0;
+        for bound in &bounds {
+            if bound >= &loc {
+                return last;
+            }
+            last = *bound;
+        }
+        *bounds.last().unwrap_or(&0)
+    }
+
     /// Render part of the row
     /// When trying to render X axis offset, this is the ideal function to use
     /// ```ignore
@@ -161,19 +200,22 @@ impl Row {
     #[must_use]
     pub fn render(&self, range: std::ops::RangeFrom<usize>) -> String {
         let mut start = range.start;
-        let text = self.render_full();
+        // Render the row
+        let text = self.render_raw();
+        let tab_width = self.get_tab_width();
         // Return an empty string if start is out of range
-        if start >= text.width() {
+        if start >= width(&text, tab_width) {
             return st!("");
         }
         // Obtain the character indices
-        let ind = raw_indices(&text, &self.indices);
+        let ind = raw_indices(&text, &self.indices, tab_width);
         // Shift the cut point forward until on a character boundary
         let space = !ind.contains_key(&start);
         while !ind.contains_key(&start) {
             start += 1;
         }
         // Perform cut and format
+        let text = text.replace("\t", &" ".repeat(tab_width));
         format!("{}{}", if space { " " } else { "" }, &text[ind[&start]..])
     }
 
@@ -235,13 +277,7 @@ impl Row {
         let mut data = vec![&'\x00'];
         data.splice(1.., text);
         data.iter()
-            .map(|c| {
-                if c == &&'\t' {
-                    tab_width
-                } else {
-                    c.width().unwrap_or(0)
-                }
-            })
+            .map(|c| width_char(**c, tab_width))
             .scan(0, |a, x| {
                 *a += x;
                 Some(*a)
